//----------------------------------------------------------------------
//  file : mul4x4b.s
//  2017/01/27 Jun Mizutani (http://www.mztn.org/)
//----------------------------------------------------------------------

.ifndef __M4x4MUL
__M4x4MUL = 1

.text
//----------------------------------------------------------------------
// [c:x0] = [a:x1] * [b:x2]
// [a]  v1= m0, m1   v2= m2, m3   v3= m4, m5   v4= m6, m7
//      v5= m8, m9   v6=m10,m11   v7=m12,m13   v8=m14,m15
// [b] v11= m0, m1  v12= m2, m3  v13= m4, m5  v14= m6, m7
//     v15= m8, m9  v16=m10,m11  v17=m12,m13  v18=m14,m15
// [c] v21= m0, m1  v22= m2, m3  v23= m4, m5  v24= m6, m7
//     v25= m8, m9  v26=m10,m11  v27=m12,m13  v28=m14,m15
// M4x4= m0, m1, m2, m3, m4, m5, m6, m7, m8, m9,m10,m11,m12,m13,m14,m15
//        x,  0,  0,  0,  0,  y,  0,  0,  0,  0,  z,  0, tx, ty, tz,  1
//----------------------------------------------------------------------
mul4x4b:

        stp     x0, x30, [sp, #-16]!
        stp     x1, x2,  [sp, #-16]!
        stp     q1, q2,  [sp, #-32]!    // save v1.2d, v2.2d
        stp     q3, q4,  [sp, #-32]!    // save v3.2d, v4.2d
        stp     q5, q6,  [sp, #-32]!    // save v5.2d, v6.2d
        stp     q7, q8,  [sp, #-32]!    // save v7.2d, v8.2d
        stp     q11,q12, [sp, #-32]!    // save v11.2d, v12.2d
        stp     q13,q14, [sp, #-32]!    // save v13.2d, v14.2d
        stp     q15,q16, [sp, #-32]!    // save v15.2d, v16.2d
        stp     q17,q18, [sp, #-32]!    // save v17.2d, v18.2d
        stp     q21,q22, [sp, #-32]!    // save v21.2d, v22.2d
        stp     q23,q24, [sp, #-32]!    // save v23.2d, v24.2d
        stp     q25,q26, [sp, #-32]!    // save v25.2d, v26.2d
        stp     q27,q28, [sp, #-32]!    // save v27.2d, v28.2d

        ld1 {v1.2d, v2.2d, v3.2d, v4.2d},[x1],#64  // a0 .. a7
        ld1 {v5.2d, v6.2d, v7.2d, v8.2d},[x1]      // a8 .. a15
        ld1 {v11.2d,v12.2d,v13.2d,v14.2d},[x2],#64 // b0 .. b7
        ld1 {v15.2d,v16.2d,v17.2d,v18.2d},[x2]     // b8 .. b15

        fmul v21.2d, v1.2d, v11.2d[0] //  c.m0 =a.m0 * b.m0,  c.m1 =a.m1 *b.m0
        fmul v22.2d, v2.2d, v11.2d[0] //  c.m2 =a.m2 * b.m0,  c.m3 =a.m3 *b.m0
        fmul v23.2d, v1.2d, v13.2d[0] //  c.m4 = a.m0 *b.m4,  c.m5 =a.m1 * b.m4
        fmul v24.2d, v2.2d, v13.2d[0] //  c.m6 =a.m2 * b.m4,  c.m7 =a.m3 * b.m4
        fmul v25.2d, v1.2d, v15.2d[0] //  c.m8 =a.m0 * b.m8,  c.m9 =a.m1 * b.m8
        fmul v26.2d, v2.2d, v15.2d[0] // c.m10 =a.m2 *b.m12, c.m11 =a.m3 *b.m12
        fmul v27.2d, v1.2d, v17.2d[0] // c.m12 =a.m0 * b.m8, c.m13 =a.m1 * b.m8
        fmul v28.2d, v2.2d, v17.2d[0] // c.m14 =a.m2 *b.m12, c.m15 =a.m3 *b.m12

        fmla v21.2d, v3.2d, v11.2d[1] //  c.m0+=a.m4 * b.m1,  c.m1+=a.m5 *b.m1
        fmla v22.2d, v4.2d, v11.2d[1] //  c.m2+=a.m6 * b.m1,  c.m3+=a.m7 *b.m1
        fmla v23.2d, v3.2d, v13.2d[1] //  c.m4+= a.m4 *b.m5,  c.m5+=a.m5 * b.m5
        fmla v24.2d, v4.2d, v13.2d[1] //  c.m6+=a.m6 * b.m5,  c.m7+=a.m7 * b.m5
        fmla v25.2d, v3.2d, v15.2d[1] //  c.m8+=a.m4 * b.m9,  c.m9+=a.m5 * b.m9
        fmla v26.2d, v4.2d, v15.2d[1] // c.m10+=a.m6 *b.m13, c.m11+=a.m7 *b.m13
        fmla v27.2d, v3.2d, v17.2d[1] // c.m12+=a.m4 * b.m9, c.m13+=a.m5 * b.m9
        fmla v28.2d, v4.2d, v17.2d[1] // c.m14+=a.m6 *b.m13, c.m15+=a.m7 *b.m13

        fmla v21.2d, v5.2d, v12.2d[0] //  c.m0+=a.m8 * b.m2,  c.m1+=a.m9 *b.m2
        fmla v22.2d, v6.2d, v12.2d[0] //  c.m2+=a.m10* b.m2,  c.m3+=a.m11*b.m2
        fmla v23.2d, v5.2d, v14.2d[0] //  c.m4+= a.m8 *b.m6,  c.m5+=a.m9 * b.m6
        fmla v24.2d, v6.2d, v14.2d[0] //  c.m6+=a.m10* b.m6,  c.m7+=a.m11* b.m6
        fmla v25.2d, v5.2d, v16.2d[0] //  c.m8+=a.m8 *b.m10,  c.m9+=a.m9 *b.m10
        fmla v26.2d, v6.2d, v16.2d[0] // c.m10+=a.m10*b.m14, c.m11+=a.m11*b.m14
        fmla v27.2d, v5.2d, v18.2d[0] // c.m12+=a.m8 *b.m10, c.m13+=a.m9 *b.m10
        fmla v28.2d, v6.2d, v18.2d[0] // c.m14+=a.m10*b.m14, c.m15+=a.m11*b.m14

        fmla v21.2d, v7.2d, v12.2d[1] //  c.m0+=a.m12* b.m3,  c.m1+=a.m13*b.m3
        fmla v22.2d, v8.2d, v12.2d[1] //  c.m2+=a.m14* b.m3,  c.m3+=a.m15*b.m3
        fmla v23.2d, v7.2d, v14.2d[1] //  c.m4+= a.m12*b.m7,  c.m5+=a.m13* b.m7
        fmla v24.2d, v8.2d, v14.2d[1] //  c.m6+=a.m14* b.m7,  c.m7+=a.m15* b.m7
        fmla v25.2d, v7.2d, v16.2d[1] //  c.m8+=a.m12*b.m11,  c.m9+=a.m13*b.m11
        fmla v26.2d, v8.2d, v16.2d[1] // c.m10+=a.m14*b.m15, c.m11+=a.m15*b.m15
        fmla v27.2d, v7.2d, v18.2d[1] // c.m12+=a.m12*b.m11, c.m13+=a.m13*b.m11
        fmla v28.2d, v8.2d, v18.2d[1] // c.m14+=a.m14*b.m15, c.m15+=a.m15*b.m15

        st1     {v21.2d,v22.2d,v23.2d,v24.2d},[x0],#64
        st1     {v25.2d,v26.2d,v27.2d,v28.2d},[x0]
        ldp     q27, q28,  [sp], #32  // restore v27.2d, v28.2d
        ldp     q25, q26,  [sp], #32  // restore v25.2d, v26.2d
        ldp     q23, q24,  [sp], #32  // restore v23.2d, v24.2d
        ldp     q21, q22,  [sp], #32  // restore v21.2d, v22.2d
        ldp     q17, q18,  [sp], #32  // restore v17.2d, v18.2d
        ldp     q15, q16,  [sp], #32  // restore v15.2d, v16.2d
        ldp     q13, q14,  [sp], #32  // restore v13.2d, v14.2d
        ldp     q11, q12,  [sp], #32  // restore v11.2d, v12.2d
        ldp     q7, q8,  [sp], #32    // restore v7.2d, v8.2d
        ldp     q5, q6,  [sp], #32    // restore v5.2d, v6.2d
        ldp     q3, q4,  [sp], #32    // restore v3.2d, v4.2d
        ldp     q1, q2,  [sp], #32    // restore v1.2d, v2.2d
        ldp     x1, x2,  [sp], #16
        ldp     x0, x30, [sp], #16
        ret

.endif
